@@ -31,7 +31,7 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 static char default_suffixes[]
   = ".out .a .ln .o .c .cc .C .p .f .F .r .y .l .s .S \
 .mod .sym .def .h .info .dvi .tex .texinfo .texi .txinfo \
-.cweb .web .sh .elc .el";
+.w .ch .web .sh .elc .el";
 
 static struct pspec default_pattern_rules[] =
   {
@@ -44,6 +44,12 @@ static struct pspec default_pattern_rules[] =
     { "%.out", "%",
 	"@rm -f $@ \n cp $< $@" },
 
+    /* Syntax is "ctangle foo.w foo.ch foo.c".  */
+    { "%.c", "%.w %.ch",
+	"$(CTANGLE) $^ $@" },
+    { "%.tex", "%.w %.ch",
+	"$(CWEAVE) $^ $@" },
+
     { 0, 0, 0 }
   };
 
@@ -171,14 +177,14 @@ static char *default_suffix_rules[] =
     ".txinfo.dvi",
     "$(TEXI2DVI) $<",
 
-    ".cweb.c",
-    "$(CTANGLE) $<",
+    ".w.c",
+    "$(CTANGLE) $< - $@",	/* The `-' says there is no `.ch' file.  */
 
     ".web.p",
     "$(TANGLE) $<",
 
-    ".cweb.tex",
-    "$(CWEAVE) $<",
+    ".w.tex",
+    "$(CWEAVE) $< - $@",	/* The `-' says there is no `.ch' file.  */
 
     ".web.tex",
     "$(WEAVE) $<",
@@ -305,8 +311,8 @@ set_default_suffixes ()
     {
       char *p = default_suffixes;
       suffix_file->deps = (struct dep *)
-	multi_glob (parse_file_seq (&p, '\0', sizeof (struct dep)),
-		    sizeof (struct dep), 1);
+	multi_glob (parse_file_seq (&p, '\0', sizeof (struct dep), 1),
+		    sizeof (struct dep));
       (void) define_variable ("SUFFIXES", 8, default_suffixes, o_default, 0);
     }
 }
