@@ -31,6 +31,7 @@
 				the nlist n_name element is a pointer,
 				not an array.
    NLIST_NAME_UNION		struct nlist has an n_un member, not n_name.
+   LINUX_LDAV_FILE		[LINUX]: Name of file containing load averages.
 
    Specific system predefines this file uses, aside from setting
    default values if not emacs:
@@ -48,6 +49,8 @@
    UMAX
    UMAX4_3
    VMS
+   LINUX			Linux: assumes /proc filesystem mounted.
+   				Support from Michael K. Johnson.
 
    In addition, to avoid nesting many #ifdefs, we internally set
    LDAV_DONE to indicate that the load average has been computed.
@@ -121,10 +124,15 @@
 #define SUNOS_5
 #endif
 
-#if defined (__osf__) && defined (__alpha__)
+#if defined (__osf__) && (defined (__alpha) || defined (__alpha__))
 #define OSF_ALPHA
 #endif
 
+#if defined (__osf__) && (defined (mips) || defined (__mips__))
+#define OSF_MIPS
+#include <sys/table.h>
+#endif
+
 /* UTek's /bin/cc on the 4300 has no architecture specific cpp define by
    default, but _MACH_IND_SYS_TYPES is defined in <sys/types.h>.  Combine
    that with a couple of other things and we'll have a unique match.  */
@@ -133,11 +141,54 @@
 #endif
 
 
-/* VAX C can't handle multi-line #ifs.  */
-#if (defined(MORE_BSD) || defined(sun) || defined(decstation) || defined(_SEQUENT_) || defined(sgi) || defined(SVR4) || defined(sony_news) || defined(sequent) || defined (OSF_ALPHA) || (defined (ardent) && defined (titan)) || defined (tek4300))
+/* VAX C can't handle multi-line #ifs, or lines longer than 256 chars.  */
+#ifndef LOAD_AVE_TYPE
+
+#ifdef MORE_BSD
 #define LOAD_AVE_TYPE long
 #endif
 
+#ifdef sun
+#define LOAD_AVE_TYPE long
+#endif
+
+#ifdef decstation
+#define LOAD_AVE_TYPE long
+#endif
+
+#ifdef _SEQUENT_
+#define LOAD_AVE_TYPE long
+#endif
+
+#ifdef sgi
+#define LOAD_AVE_TYPE long
+#endif
+
+#ifdef SVR4
+#define LOAD_AVE_TYPE long
+#endif
+
+#ifdef sony_news
+#define LOAD_AVE_TYPE long
+#endif
+
+#ifdef sequent
+#define LOAD_AVE_TYPE long
+#endif
+
+#ifdef OSF_ALPHA
+#define LOAD_AVE_TYPE long
+#endif
+
+#if defined (ardent) && defined (titan)
+#define LOAD_AVE_TYPE long
+#endif
+
+#ifdef tek4300
+#define LOAD_AVE_TYPE long
+#endif
+
+#endif /* No LOAD_AVE_TYPE.  */
 
 #ifndef	FSCALE
 
@@ -152,6 +203,9 @@
 #endif
 
 #if defined (sgi) || defined (sequent)
+/* Sometimes both MIPS and sgi are defined, so FSCALE was just defined
+   above under #ifdef MIPS.  But we want the sgi value.  */
+#undef FSCALE
 #define	FSCALE 1000.0
 #endif
 
@@ -169,11 +223,59 @@
 #define	LDAV_CVT(n) (((double) (n)) / FSCALE)
 #endif
 
-/* VAX C can't handle multi-line #ifs.  */
-#if !defined(NLIST_STRUCT) && (defined(MORE_BSD) || defined(sun) || defined(decstation) || defined(hpux) || defined(_SEQUENT_) || defined(sequent) || defined(sgi) || defined(SVR4) || defined(sony_news) || defined (OSF_ALPHA) || (defined (ardent) && defined (titan)) || defined (tek4300))
+/* VAX C can't handle multi-line #ifs, or lines longer that 256 characters.  */
+#ifndef NLIST_STRUCT
+
+#ifdef MORE_BSD
 #define NLIST_STRUCT
 #endif
 
+#ifdef sun
+#define NLIST_STRUCT
+#endif
+
+#ifdef decstation
+#define NLIST_STRUCT
+#endif
+
+#ifdef hpux
+#define NLIST_STRUCT
+#endif
+
+#if defined (_SEQUENT_) || defined (sequent)
+#define NLIST_STRUCT
+#endif
+
+#ifdef sgi
+#define NLIST_STRUCT
+#endif
+
+#ifdef SVR4
+#define NLIST_STRUCT
+#endif
+
+#ifdef sony_news
+#define NLIST_STRUCT
+#endif
+
+#ifdef OSF_ALPHA
+#define NLIST_STRUCT
+#endif
+
+#if defined (ardent) && defined (titan)
+#define NLIST_STRUCT
+#endif
+
+#ifdef tex4300
+#define NLIST_STRUCT
+#endif
+
+#ifdef butterfly
+#define NLIST_STRUCT
+#endif
+
+#endif /* defined (NLIST_STRUCT) */
+
 
 #if defined(sgi) || (defined(mips) && !defined(BSD))
 #define FIXUP_KERNEL_SYMBOL_ADDR(nl) ((nl)[0].n_value &= ~(1 << 31))
@@ -337,7 +439,7 @@ static kvm_t *kd;
 
 /* Put the 1 minute, 5 minute and 15 minute load averages
    into the first NELEM elements of LOADAVG.
-   Return the number written (never more than 3),
+   Return the number written (never more than 3, but may be less than NELEM),
    or -1 if an error occurred.  */
 
 int
@@ -347,6 +449,46 @@ getloadavg (loadavg, nelem)
 {
   int elem = 0;			/* Return value.  */
 
+#ifdef NO_GET_LOAD_AVG
+#define LDAV_DONE
+  /* Set errno to zero to indicate that there was no particular error;
+     this function just can't work at all on this system.  */
+  errno = 0;
+  elem = -1;
+#endif
+
+#if !defined (LDAV_DONE) && defined (LINUX)
+#define LDAV_DONE
+#undef LOAD_AVE_TYPE
+
+#ifndef LINUX_LDAV_FILE
+#define LINUX_LDAV_FILE "/proc/loadavg"
+#endif
+
+  char ldavgbuf[40];
+  double load_ave[3];
+  int fd, count;
+
+  fd = open (LINUX_LDAV_FILE, O_RDONLY);
+  if (fd == -1)
+    return -1;
+  count = read (fd, ldavgbuf, 40);
+  (void) close (fd);
+  if (count <= 0)
+    return -1;
+
+  count = sscanf (ldavgbuf, "%lf %lf %lf",
+		  &load_ave[0], &load_ave[1], &load_ave[2]);
+  if (count < 1)
+    return -1;
+
+  for (elem = 0; elem < nelem && elem < count; elem++)
+    loadavg[elem] = load_ave[elem];
+
+  return elem;
+
+#endif /* LINUX */
+
 #if !defined (LDAV_DONE) && defined (NeXT)
 #define LDAV_DONE
   /* The NeXT code was adapted from iscreen 3.2.  */
@@ -355,12 +497,8 @@ getloadavg (loadavg, nelem)
   struct processor_set_basic_info info;
   unsigned info_count;
 
-  if (nelem > 1)
-    {
-      /* We only know how to get the 1-minute average for this system.  */
-      errno = EINVAL;
-      return -1;
-    }
+  /* We only know how to get the 1-minute average for this system,
+     so even if the caller asks for more than 1, we only return 1.  */
 
   if (!getloadavg_initialized)
     {
@@ -498,6 +636,17 @@ getloadavg (loadavg, nelem)
     loadavg[elem++] = load_ave[2] / 65536.0;
 #endif /* apollo */
 
+#if !defined (LDAV_DONE) && defined (OSF_MIPS)
+#define LDAV_DONE
+
+  struct tbl_loadavg load_ave;
+  table (TBL_LOADAVG, 0, &load_ave, 1, sizeof (load_ave));
+  loadavg[elem++]
+    = (load_ave.tl_lscale == 0
+       ? load_ave.tl_avenrun.d[0]
+       : (load_ave.tl_avenrun.l[0] / (double) load_ave.tl_lscale));
+#endif	/* OSF_MIPS */
+
 #if !defined (LDAV_DONE) && defined (VMS)
   /* VMS specific code -- read from the Load Ave driver.  */
 
@@ -549,7 +698,6 @@ getloadavg (loadavg, nelem)
   /* Get the address of LDAV_SYMBOL.  */
   if (offset == 0)
     {
-#ifndef SUNOS_5
 #ifndef sgi
 #ifndef NLIST_STRUCT
       strcpy (nl[0].n_name, LDAV_SYMBOL);
@@ -564,6 +712,7 @@ getloadavg (loadavg, nelem)
 #endif /* not NLIST_NAME_UNION */
 #endif /* NLIST_STRUCT */
 
+#ifndef SUNOS_5
       if (nlist (KERNEL_FILE, nl) >= 0)
 	/* Omit "&& nl[0].n_type != 0 " -- it breaks on Sun386i.  */
 	{
@@ -572,6 +721,7 @@ getloadavg (loadavg, nelem)
 #endif
 	  offset = nl[0].n_value;
 	}
+#endif  /* !SUNOS_5 */
 #else /* sgi */
       int ldav_off;
 
@@ -579,7 +729,6 @@ getloadavg (loadavg, nelem)
       if (ldav_off != -1)
 	offset = (long) ldav_off & 0x7fffffff;
 #endif /* sgi */
-#endif  /* !SUNOS_5 */
     }
 
   /* Make sure we have /dev/kmem open.  */
@@ -590,10 +739,14 @@ getloadavg (loadavg, nelem)
       if (channel >= 0)
 	getloadavg_initialized = 1;
 #else /* SUNOS_5 */
+      /* We pass 0 for the kernel, corefile, and swapfile names
+	 to use the currently running kernel.  */
       kd = kvm_open (0, 0, 0, O_RDONLY, 0);
       if (kd != 0) 
 	{
+	  /* nlist the currently running kernel.  */
 	  kvm_nlist (kd, nl);
+	  offset = nl[0].n_value;
 	  getloadavg_initialized = 1;
 	}
 #endif /* SUNOS_5 */
