@@ -113,7 +113,8 @@ pattern_search (file, archive, depth, recursions)
 
   /* Buffer in which we store all the rules that are possibly applicable.  */
   struct rule **tryrules
-    = (struct rule **) alloca (num_pattern_rules * sizeof (struct rule *));
+    = (struct rule **) alloca (num_pattern_rules * max_pattern_targets
+			       * sizeof (struct rule *));
 
   /* Number of valid elements in TRYRULES.  */
   unsigned int nrules;
@@ -164,9 +165,6 @@ pattern_search (file, archive, depth, recursions)
   nrules = 0;
   for (rule = pattern_rules; rule != 0; rule = rule->next)
     {
-      int specific_rule_may_have_matched = 0;
-      int check_lastslash;
-
       /* If the pattern rule has deps but no commands, ignore it.
 	 Users cancel built-in rules by redefining them without commands.  */
       if (rule->deps != 0 && rule->cmds == 0)
@@ -184,6 +182,7 @@ pattern_search (file, archive, depth, recursions)
 	{
 	  char *target = rule->targets[i];
 	  char *suffix = rule->suffixes[i];
+	  int check_lastslash;
 
 	  /* Rules that can match any filename and are not terminal
 	     are ignored if we're recursing, so that they cannot be
@@ -237,29 +236,21 @@ pattern_search (file, archive, depth, recursions)
 
 	  /* Record if we match a rule that not all filenames will match.  */
 	  if (target[1] != '\0')
-	    specific_rule_may_have_matched = 1;
+	    specific_rule_matched = 1;
 
-	  /* We have a matching target.  Don't search for any more.  */
-	  break;
+	  /* A rule with no dependencies and no commands exists solely to set
+	     specific_rule_matched when it matches.  Don't try to use it.  */
+	  if (rule->deps == 0 && rule->cmds == 0)
+	    continue;
+
+	  /* Record this rule in TRYRULES and the index of the matching
+	     target in MATCHES.  If several targets of the same rule match,
+	     that rule will be in TRYRULES more than once.  */
+	  tryrules[nrules] = rule;
+	  matches[nrules] = i;
+	  checked_lastslash[nrules] = check_lastslash;
+	  ++nrules;
 	}
-
-      /* None of the targets matched.  */
-      if (rule->targets[i] == 0)
-	continue;
-
-      specific_rule_matched |= specific_rule_may_have_matched;
-
-      /* A rule with no dependencies and no commands exists solely to set
-	 specific_rule_matched when it matches.  Don't try to use it.  */
-      if (rule->deps == 0 && rule->cmds == 0)
-	continue;
-
-      /* Record this rule in TRYRULES and the index
-	 of the (first) matching target in MATCHES.  */
-      tryrules[nrules] = rule;
-      matches[nrules] = i;
-      checked_lastslash[nrules] = check_lastslash;
-      ++nrules;
     }
 
   /* If we have found a matching rule that won't match all filenames,
@@ -308,8 +299,7 @@ pattern_search (file, archive, depth, recursions)
 	  stem = filename
 	    + (rule->suffixes[matches[i]] - rule->targets[matches[i]]) - 1;
 	  stemlen = namelen - rule->lens[matches[i]] + 1;
-	  check_lastslash = (lastslash != 0
-			     && index (rule->targets[matches[i]], '/') == 0);
+	  check_lastslash = checked_lastslash[i];
 	  if (check_lastslash)
 	    {
 	      stem += lastslash - filename + 1;
@@ -331,6 +321,7 @@ pattern_search (file, archive, depth, recursions)
 		  register unsigned int i;
 		  if (check_lastslash)
 		    {
+		      /* Copy directory name from the original FILENAME.  */
 		      i = lastslash - filename + 1;
 		      bcopy (filename, depname, i);
 		    }
@@ -354,7 +345,7 @@ pattern_search (file, archive, depth, recursions)
 		     "impossible", then the rule fails and don't
 		     bother trying it on the second pass either
 		     since we know that will fail too.  */
-		  DEBUGP2 ("Rejecting impossible %s dependent `%s'.\n",
+		  DEBUGP2 ("Rejecting impossible %s dependency `%s'.\n",
 			   p == depname ? "implicit" : "rule", p);
 		  tryrules[i] = 0;
 		  break;
@@ -364,14 +355,23 @@ pattern_search (file, archive, depth, recursions)
 
 	      DEBUGP2 ("Trying %s dependency `%s'.\n",
 		       p == depname ? "implicit" : "rule", p);
-	      if (!rule->subdir && (lookup_file (p) != 0 || file_exists_p (p)))
+
+	      /* The DEP->changed flag says that this dependency resides in a
+		 nonexistent directory.  So we normally can skip looking for
+		 the file.  However, if CHECK_LASTSLASH is set, then the
+		 dependency file we are actually looking for is in a different
+		 directory (the one gotten by prepending FILENAME's directory),
+		 so it might actually exist.  */
+
+	      if ((!dep->changed || check_lastslash)
+		  && (lookup_file (p) != 0 || file_exists_p (p)))
 		{
 		  found_files[deps_found++] = savestring (p, strlen (p));
 		  continue;
 		}
 	      /* This code, given FILENAME = "lib/foo.o", dependency name
 		 "lib/foo.c", and VPATH=src, searches for "src/lib/foo.c".  */
-	      if (vpath_search (&p))
+	      if (vpath_search (&p, (time_t *) 0))
 		{
 		  DEBUGP2 ("Found dependency as `%s'.%s\n", p, "");
 		  found_files[deps_found++] = p;
@@ -399,10 +399,13 @@ pattern_search (file, archive, depth, recursions)
 		      p = savestring (p, strlen (p));
 		      intermediate_patterns[deps_found]
 			= intermediate_file->name;
-		      found_files[deps_found] = p;
 		      intermediate_file->name = p;
 		      intermediate_files[deps_found] = intermediate_file;
 		      intermediate_file = 0;
+		      /* Allocate an extra copy to go in FOUND_FILES,
+			 because every elt of FOUND_FILES is consumed
+			 or freed later.  */
+		      found_files[deps_found] = savestring (p, strlen (p));
 		      ++deps_found;
 		      continue;
 		    }
@@ -504,7 +507,14 @@ pattern_search (file, archive, depth, recursions)
       if (recursions == 0)
 	{
 	  dep->name = 0;
-	  dep->file = enter_file (s);	
+	  dep->file = lookup_file (s);
+	  if (dep->file == 0)
+	    /* enter_file consumes S's storage.  */
+	    dep->file = enter_file (s);
+	  else
+	    /* A copy of S is already allocated in DEP->file->name.
+	       So we can free S.  */
+	    free (s);
 	}
       else
 	{
@@ -532,6 +542,8 @@ pattern_search (file, archive, depth, recursions)
     file->stem = stem[stemlen] == '\0' ? stem : savestring (stem, stemlen);
   else
     {
+      /* We want to prepend the directory from
+	 the original FILENAME onto the stem.  */
       file->stem = (char *) xmalloc (((lastslash + 1) - filename)
 				     + stemlen + 1);
       bcopy (filename, file->stem, (lastslash + 1) - filename);
