@@ -975,15 +975,15 @@ search_path (file, path, program)
 #endif
       static int ngroups = -1;
 #ifdef	NGROUPS_MAX
-      static gid_t groups[NGROUPS_MAX];
+      static GETGROUPS_T groups[NGROUPS_MAX];
 #define	ngroups_max	NGROUPS_MAX
 #else
-      static gid_t *groups = 0;
+      static GETGROUPS_T *groups = 0;
       static int ngroups_max;
       if (groups == 0)
 	{
 	  ngroups_max = GET_NGROUPS_MAX;
-	  groups = (gid_t *) malloc (ngroups_max * sizeof (gid_t));
+	  groups = (GETGROUPS_T *) malloc (ngroups_max * sizeof (GETGROUPS_T));
 	}
 #endif
       if (groups != 0 && ngroups == -1)
@@ -1390,10 +1390,20 @@ construct_command_argv (line, restp, file)
      char *line, **restp;
      struct file *file;
 {
-  char *shell = allocated_variable_expand_for_file ("$(SHELL)", file);
-  char *ifs = allocated_variable_expand_for_file ("$(IFS)", file);
+  char *shell, *ifs;
   char **argv;
 
+  {
+    /* Turn off --warn-undefined-variables while we expand SHELL and IFS.  */
+    int save = warn_undefined_variables_flag;
+    warn_undefined_variables_flag = 0;
+
+    shell = allocated_variable_expand_for_file ("$(SHELL)", file);
+    ifs = allocated_variable_expand_for_file ("$(IFS)", file);
+
+    warn_undefined_variables_flag = save;
+  }
+
   argv = construct_command_argv_internal (line, restp, shell, ifs);
 
   free (shell);
