@@ -26,8 +26,6 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 extern char *version_string;
 
-extern struct dep *read_all_makefiles ();
-
 extern void print_variable_data_base ();
 extern void print_dir_data_base ();
 extern void print_rule_data_base ();
@@ -49,35 +47,7 @@ static void log_working_directory ();
 static void print_data_base (), print_version ();
 static void decode_switches (), decode_env_switches ();
 static void define_makeflags ();
-
-
-#if 0 /* dummy tag */
-flags () {}
-#endif
-/* Flags:
- *	-b ignored for compatibility with System V Make
- *	-C change directory
- *	-d debug
- *	-e env_overrides
- *	-f makefile
- *	-i ignore_errors
- *	-j job_slots
- *	-k keep_going
- *	-l max_load_average
- *	-m ignored for compatibility with something or other
- *	-n just_print
- *	-o consider file old
- *	-p print_data_base
- *	-q question
- *	-r no_builtin_rules
- *	-s silent
- *	-S turn off -k
- *	-t touch
- *	-v print version information
- *	-w log working directory
- *	-W consider file new (with -n, `what' if effect)
- */
-
+
 /* The structure that describes an accepted command switch.  */
 
 struct command_switch
@@ -203,7 +173,7 @@ static unsigned int inf_jobs = 0;
 double max_load_average = -1.0;
 double default_load_average = -1.0;
 
-/* List of directories given with -c switches.  */
+/* List of directories given with -C switches.  */
 
 static struct stringlist *directories = 0;
 
@@ -222,6 +192,11 @@ static struct stringlist *new_files = 0;
 /* If nonzero, we should just print usage and exit.  */
 
 static int print_usage_flag = 0;
+
+/* If nonzero, we should print a warning message
+   for each reference to an undefined variable.  */
+
+int warn_undefined_variables_flag;
 
 /* The table of command switches.  */
 
@@ -303,6 +278,9 @@ static const struct command_switch switches[] =
     { 'W', string, (char *) &new_files, 0, 0, 0, 0, 0,
 	"what-if", "FILE",
 	"Consider FILE to be infinitely new" },
+    { 2, flag, (char *) &warn_undefined_variables_flag, 1, 1, 0, 0, 0,
+	"warn-undefined-variables", 0,
+	"Warn when an undefined variable is referenced" },
     { '\0', }
   };
 
@@ -365,6 +343,41 @@ int fatal_signal_mask;
 #endif
 #endif
 
+static struct file *
+enter_command_line_file (name)
+     char *name;
+{
+  if (name[0] == '~')
+    {
+      char *expanded = tilde_expand (name);
+      if (expanded != 0)
+	name = expanded;	/* Memory leak; I don't care.  */
+    }
+
+  /* This is also done in parse_file_seq, so this is redundant
+     for names read from makefiles.  It is here for names passed
+     on the command line.  */
+  while (name[0] == '.' && name[1] == '/' && name[2] != '\0')
+    {
+      name += 2;
+      while (*name == '/')
+	/* Skip following slashes: ".//foo" is "foo", not "/foo".  */
+	++name;
+    }
+  
+  if (*name == '\0')
+    {
+      /* It was all slashes!  Move back to the dot and truncate
+	 it after the first slash, so it becomes just "./".  */
+      do
+	--name;
+      while (name[0] != '.');
+      name[2] = '\0';
+    }
+
+  return enter_file (savestring (name, strlen (name)));
+}
+
 int
 main (argc, argv, envp)
      int argc;
@@ -550,7 +563,7 @@ main (argc, argv, envp)
 	{
 	  /* It was not a variable definition, so it is a target to be made.
 	     Enter it as a file and add it to the dep chain of goals.  */
-	  f = enter_file (other_args->list[i]);
+	  f = enter_command_line_file (other_args->list[i]);
 	  f->cmd_target = 1;
 	  
 	  if (goals == 0)
@@ -599,8 +612,16 @@ main (argc, argv, envp)
     for (i = 0; directories->list[i] != 0; ++i)
       {
 	char *dir = directories->list[i];
+	if (dir[0] == '~')
+	  {
+	    char *expanded = tilde_expand (dir);
+	    if (expanded != 0)
+	      dir = expanded;
+	  }
 	if (chdir (dir) < 0)
 	  pfatal_with_name (dir);
+	if (dir != directories->list[i])
+	  free (dir);
       }
 
   /* Figure out the level of recursion.  */
@@ -663,9 +684,6 @@ main (argc, argv, envp)
 	    static char name[] = "/tmp/GmXXXXXX";
 	    FILE *outfile;
 
-	    /* Free the storage allocated for "-".  */
-	    free (makefiles->list[i]);
-
 	    /* Make a unique filename.  */
 	    (void) mktemp (name);
 
@@ -686,7 +704,13 @@ main (argc, argv, envp)
 
 	    /* Replace the name that read_all_makefiles will
 	       see with the name of the temporary file.  */
-	    makefiles->list[i] = savestring (name, sizeof name - 1);
+	    {
+	      char *temp;
+	      /* SGI compiler requires alloca's result be assigned simply.  */
+	      temp = (char *) alloca (sizeof (name));
+	      bcopy (name, temp, sizeof (name));
+	      makefiles->list[i] = temp;
+	    }
 
 	    /* Make sure the temporary file will not be remade.  */
 	    f = enter_file (savestring (name, sizeof name - 1));
@@ -776,7 +800,7 @@ main (argc, argv, envp)
   if (old_files != 0)
     for (p = old_files->list; *p != 0; ++p)
       {
-	f = enter_file (*p);
+	f = enter_command_line_file (*p);
 	f->last_mtime = (time_t) 1;
 	f->updated = 1;
 	f->update_status = 0;
@@ -788,7 +812,7 @@ main (argc, argv, envp)
       now = time ((time_t *) 0);
       for (p = new_files->list; *p != 0; ++p)
 	{
-	  f = enter_file (*p);
+	  f = enter_command_line_file (*p);
 	  f->last_mtime = now;
 	}
     }
@@ -894,7 +918,7 @@ main (argc, argv, envp)
 			any_remade |= (file_mtime_no_search (d->file)
 				       != makefile_mtimes[i]);
 		      }
-		    else if (d->changed != 1)
+		    else if (! (d->changed & RM_DONTCARE))
 		      {
 			time_t mtime;
 			/* The update failed and this makefile was not
@@ -908,23 +932,20 @@ main (argc, argv, envp)
 		  }
 		else
 		  /* This makefile was not found at all.  */
-		  switch (d->changed)
+		  if (! (d->changed & RM_DONTCARE))
 		    {
-		    case 0:
-		      /* A normal makefile.  We must die later.  */
-		      error ("Makefile `%s' was not found", dep_name (d));
-		      any_failed = 1;
-		      break;
-		    case 1:
-		      /* A makefile from the MAKEFILES variable.
-			 We don't care.  */
-		      break;
-		    case 2:
-		      /* An included makefile.  We don't need
-			 to die, but we do want to complain.  */
-		      error ("Included makefile `%s' was not found.",
-			     dep_name (d));
-		      break;
+		      /* This is a makefile we care about.  See how much.  */
+		      if (d->changed & RM_INCLUDED)
+			/* An included makefile.  We don't need
+			   to die, but we do want to complain.  */
+			error ("Included makefile `%s' was not found.",
+			       dep_name (d));
+		      else
+			{
+			  /* A normal makefile.  We must die later.  */
+			  error ("Makefile `%s' was not found", dep_name (d));
+			  any_failed = 1;
+			}
 		    }
 
 		free ((char *) d);
@@ -939,18 +960,23 @@ main (argc, argv, envp)
 	  }
 
 	case 0:
-	re_exec:;
+	re_exec:
 	  /* Updated successfully.  Re-exec ourselves.  */
+
+	  remove_intermediates (0);
+
+	  if (print_data_base_flag)
+	    print_data_base ();
+
 	  if (print_directory_flag)
 	    log_working_directory (0);
-	  if (debug_flag)
-	    puts ("Re-execing myself....");
+
 	  if (makefiles != 0)
 	    {
 	      /* These names might have changed.  */
 	      register unsigned int i, j = 0;
 	      for (i = 1; i < argc; ++i)
-		if (!strcmp (argv[i], "-f"))
+		if (!strcmp (argv[i], "-f")) /* XXX */
 		  {
 		    char *p = &argv[i][2];
 		    if (*p == '\0')
@@ -960,6 +986,7 @@ main (argc, argv, envp)
 		    ++j;
 		  }
 	    }
+
 	  if (directories != 0 && directories->idx > 0)
 	    {
 	      char bad;
@@ -978,8 +1005,7 @@ main (argc, argv, envp)
 	      if (bad)
 		fatal ("Couldn't change back to original directory.");
 	    }
-	  fflush (stdout);
-	  fflush (stderr);
+
 	  for (p = environ; *p != 0; ++p)
 	    if (!strncmp (*p, "MAKELEVEL=", 10))
 	      {
@@ -992,6 +1018,19 @@ main (argc, argv, envp)
 		sprintf (*p, "MAKELEVEL=%u", makelevel);
 		break;
 	      }
+
+	  if (debug_flag)
+	    {
+	      char **p;
+	      fputs ("Re-executing:", stdout);
+	      for (p = argv; *p != 0; ++p)
+		printf (" %s", *p);
+	      puts ("");
+	    }
+
+	  fflush (stdout);
+	  fflush (stderr);
+
 	  exec_command (argv, environ);
 	  /* NOTREACHED */
 	}
@@ -1135,7 +1174,7 @@ decode_switches (argc, argv, env)
       other_args->max = argc + 1;
       other_args->list = (char **) xmalloc ((argc + 1) * sizeof (char *));
       other_args->idx = 1;
-      other_args->list[0] = savestring (argv[0], strlen (argv[0]));
+      other_args->list[0] = argv[0];
     }
 
   /* getopt does most of the parsing for us.
@@ -1205,7 +1244,7 @@ decode_switches (argc, argv, env)
 			xrealloc ((char *) sl->list,
 				  sl->max * sizeof (char *));
 		    }
-		  sl->list[sl->idx++] = savestring (optarg, strlen (optarg));
+		  sl->list[sl->idx++] = optarg;
 		  sl->list[sl->idx] = 0;
 		  break;
 
@@ -1279,7 +1318,7 @@ positive integral argument",
       fputs ("Options:\n", stderr);
       for (cs = switches; cs->c != '\0'; ++cs)
 	{
-	  char buf[1024], arg[50], *p;
+	  char buf[1024], shortarg[50], longarg[50], *p;
 
 	  if (cs->description[0] == '-')
 	    continue;
@@ -1287,13 +1326,15 @@ positive integral argument",
 	  switch (long_options[cs - switches].has_arg)
 	    {
 	    case no_argument:
-	      arg[0] = '\0';
+	      shortarg[0] = longarg[0] = '\0';
 	      break;
 	    case required_argument:
-	      sprintf (arg, " %s", cs->argdesc);
+	      sprintf (longarg, "=%s", cs->argdesc);
+	      sprintf (shortarg, " %s", cs->argdesc);
 	      break;
 	    case optional_argument:
-	      sprintf (arg, " [%s]", cs->argdesc);
+	      sprintf (longarg, "[=%s]", cs->argdesc);
+	      sprintf (shortarg, " [%s]", cs->argdesc);
 	      break;
 	    }
 
@@ -1301,7 +1342,7 @@ positive integral argument",
 
 	  if (isalnum (cs->c))
 	    {
-	      sprintf (buf, "  -%c%s", cs->c, arg);
+	      sprintf (buf, "  -%c%s", cs->c, shortarg);
 	      p += strlen (p);
 	    }
 	  if (cs->long_name != 0)
@@ -1309,14 +1350,15 @@ positive integral argument",
 	      unsigned int i;
 	      sprintf (p, "%s--%s%s",
 		       !isalnum (cs->c) ? "  " : ", ",
-		       cs->long_name, arg);
+		       cs->long_name, longarg);
 	      p += strlen (p);
 	      for (i = 0; i < (sizeof (long_option_aliases) /
 			       sizeof (long_option_aliases[0]));
 		   ++i)
 		if (long_option_aliases[i].val == cs->c)
 		  {
-		    sprintf (p, ", --%s%s", long_option_aliases[i].name, arg);
+		    sprintf (p, ", --%s%s",
+			     long_option_aliases[i].name, longarg);
 		    p += strlen (p);
 		  }
 	    }
@@ -1329,11 +1371,11 @@ positive integral argument",
 		  /* This is another switch that does the same
 		     one as the one we are processing.  We want
 		     to list them all together on one line.  */
-		  sprintf (p, ", -%c%s", ncs->c, arg);
+		  sprintf (p, ", -%c%s", ncs->c, shortarg);
 		  p += strlen (p);
 		  if (ncs->long_name != 0)
 		    {
-		      sprintf (p, ", --%s%s", ncs->long_name, arg);
+		      sprintf (p, ", --%s%s", ncs->long_name, longarg);
 		      p += strlen (p);
 		    }
 		}
@@ -1387,8 +1429,10 @@ decode_env_switches (envar, len)
     return;
 
   /* Make a copy of the value in ARGS, where we will munge it.
-     If it does not begin with a dash, prepend one.  */
-  args = (char *) alloca (1 + len + 2);
+     If it does not begin with a dash, prepend one.
+     We must allocate lasting storage for this (and we never free it) because
+     decode_switches may save pointers into it for string-valued switches.  */
+  args = (char *) xmalloc (1 + len + 2);
   if (value[0] != '-')
     args[0] = '-';
   bcopy (value, value[0] == '-' ? args : &args[1], len + 1);
