@@ -55,9 +55,10 @@ extern int errno;
 #endif
 #endif
 
-#ifdef butterfly
-/* The BBN Butterfly has <unistd.h> that defines _POSIX_VERSION,
-   but isn't really a POSIX.1 system.  */
+/* Some systems define _POSIX_VERSION but are not really POSIX.1.  */
+#if (defined (butterfly) || \
+     (defined (__mips) && defined (_SYSTYPE_SVR3)) || \
+     (defined (sequent) && defined (i386)))
 #undef POSIX
 #endif
 
@@ -121,14 +122,14 @@ extern unsigned int get_path_max ();
 #define	PATH_VAR(var)	char *var = (char *) alloca (GET_PATH_MAX)
 #endif
 
-#ifdef	uts
+#ifdef	STAT_MACROS_BROKEN
 #ifdef	S_ISREG
 #undef	S_ISREG
 #endif
 #ifdef	S_ISDIR
 #undef	S_ISDIR
 #endif
-#endif	/* uts.  */
+#endif	/* STAT_MACROS_BROKEN.  */
 
 #ifndef	S_ISREG
 #define	S_ISREG(mode)	(((mode) & S_IFMT) == S_IFREG)
@@ -306,7 +307,7 @@ extern unsigned int *reading_lineno_ptr;
 extern int just_print_flag, silent_flag, ignore_errors_flag, keep_going_flag;
 extern int debug_flag, print_data_base_flag, question_flag, touch_flag;
 extern int env_overrides, no_builtin_rules_flag, print_version_flag;
-extern int print_directory_flag;
+extern int print_directory_flag, warn_undefined_variables_flag;
 
 extern unsigned int job_slots;
 extern double max_load_average;
