@@ -389,12 +389,14 @@ define_automatic_variables ()
 int export_all_variables;
 
 /* Create a new environment for FILE's commands.
+   If FILE is nil, this is for the `shell' function.
    The child's MAKELEVEL variable is incremented.  */
 
 char **
 target_environment (file)
      struct file *file;
 {
+  struct variable_set_list *set_list;
   register struct variable_set_list *s;
   struct variable_bucket
     {
@@ -408,14 +410,19 @@ target_environment (file)
   char **result;
   unsigned int mklev_hash;
 
+  if (file == 0)
+    set_list = current_variable_set_list;
+  else
+    set_list = file->variables;
+
   /* Find the lowest number of buckets in any set in the list.  */
-  s = file->variables;
+  s = set_list;
   buckets = s->set->buckets;
   for (s = s->next; s != 0; s = s->next)
     if (s->set->buckets < buckets)
       buckets = s->set->buckets;
 
-  /* Find the hash value of `MAKELEVEL' will fall into.  */
+  /* Find the hash value of the bucket `MAKELEVEL' will fall into.  */
   {
     char *p = "MAKELEVEL";
     mklev_hash = 0;
@@ -431,7 +438,7 @@ target_environment (file)
   /* Run through all the variable sets in the list,
      accumulating variables in TABLE.  */
   nvariables = 0;
-  for (s = file->variables; s != 0; s = s->next)
+  for (s = set_list; s != 0; s = s->next)
     {
       register struct variable_set *set = s->set;
       for (i = 0; i < set->buckets; ++i)
